@@ -25,7 +25,7 @@ EE flatten_fieldaccess(EnvI& env, const Ctx& ctx, Expression* e, VarDecl* r, Var
   IntVal i = fa->field()->cast<IntLit>()->v();
   if (i < 1 || i > al->size()) {
     // This should not happen, type checking should ensure all fields are valid.
-    throw EvalError(env, fa->loc(), "Internal error: acessing invalid field");
+    throw EvalError(env, fa->loc(), "Internal error: accessing invalid field");
   }
 
   // Bind result
